/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                               field.cpp
 *
 *  Definitions of the main FieldObj concentration field class (derived from
 *  the VectorObj class), the daughter BufferObj buffer concentration field 
 *  class, and the compound buffer array class BufferArray.
 *
 *  FieldObj methods include all low-level ADI difference schemes routines.
 *  Higher-level numeric engine routines are defined in "calc.cpp"       
 *       
 *  FieldObj and BufferArray variables are member of the main SimulationObj
 *  object, defined in simulation.h / simulation.cpp
 *
 ******************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "vector.h"
#include <stdarg.h>
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "field.h"

extern void setFieldByFunction(TokenString &, long, double *, const char *errStr);                                     

const double pi = atan(1.0) * 4;

double (*integrate)(double, double, double, int) = &gaussian;

//****************************************************************************
//                      INITIALIZING STATIC MEMBERS:
//****************************************************************************

class RegionObj      *FieldObj::Region = 0;
class GridObj        *FieldObj::Grid = 0;
class BCarrayObj     *FieldObj::BCarray = 0;

char   FieldObj::SAME_CURRENT = 1;

int    FieldObj::xsize = 0,   FieldObj::ysize = 0,    FieldObj::zsize = 0;
long   FieldObj::Size = 0,    FieldObj::xysize = 0;

double *FieldObj::xgrid = 0,  *FieldObj::ygrid = 0,   *FieldObj::zgrid = 0;
double *FieldObj::xcoord = 0, *FieldObj::ycoord = 0,  *FieldObj::zcoord = 0;

double *FieldObj::dxplus = 0, *FieldObj::dxminus = 0, *FieldObj::dxinv = 0, *FieldObj::dx2inv = 0;
double *FieldObj::dyplus = 0, *FieldObj::dyminus = 0, *FieldObj::dyinv = 0, *FieldObj::dy2inv = 0;
double *FieldObj::dzplus = 0, *FieldObj::dzminus = 0, *FieldObj::dzinv = 0, *FieldObj::dz2inv = 0;
  
double *FieldObj::bc_deriv = 0, *FieldObj::bc_coef = 0, *FieldObj::bc_pump = 0, *FieldObj::bc_lin = 0;
char   **FieldObj::bc_id = 0;
int    FieldObj::bc_type_num = 0;

double *FieldObj::right = 0, *FieldObj::diag = 0, *FieldObj::sup = 0, *FieldObj::sub = 0;

nabla  FieldObj::nablaX = 0, FieldObj::nablaY = 0, FieldObj::nablaZ = 0;

struct TermStruct *Currents = 0;
class  ExpressionObj *Current = 0;

//*************************************************************************
//*                T R I D I A G O N A L   S Y S T E M
//*************************************************************************

void FieldObj::setStaticData(RegionObj &RO, GridObj &GO, BCarrayObj &BCO) {

  Region = &RO;

  Grid = &GO;
  Size = GO.Size;    xysize = GO.xysize; 
  xsize = GO.xsize;  ysize = GO.ysize; zsize = GO.zsize;

  xgrid = GO.xgrid;   ygrid = GO.ygrid;   zgrid = GO.zgrid; 
  xcoord = GO.xcoord; ycoord = GO.ycoord; zcoord = GO.zcoord; 

  dxplus = GO.dxplus;   dyplus = GO.dyplus;   dzplus = GO.dzplus; 
  dxminus = GO.dxminus; dyminus = GO.dyminus; dzminus = GO.dzminus;
  dxinv = GO.dxinv;     dyinv = GO.dyinv;     dzinv = GO.dzinv;  
  dx2inv = GO.dx2inv;   dy2inv = GO.dy2inv;   dz2inv = GO.dz2inv;  

  BCarray = &BCO;
  bc_id = BCO.bc_id;        bc_type_num = BCO.bc_type_num;
  bc_deriv = BCO.bc_deriv;  bc_coef=BCO.bc_coef;  bc_lin = BCO.bc_lin;  bc_pump = BCO.bc_pump;

  int max = ( xsize > ysize ? xsize : ysize );
  if (zsize > max) max = zsize;
  init_tridiag(max + 2);
} 


void FieldObj::init_tridiag(int n)
{
diag = new double[n];  right = new double[n];
sup = new double[n];   sub = new double[n];
}

void FieldObj::kill_tridiag()
{
  if (diag) {
    delete [] diag; delete [] right;
    delete [] sup;  delete [] sub;
  }
}

//*******************   S O L V E R :  **************************

void FieldObj::tridiag(int n)
{
int    i;
register double sb, sp, dg, rt;

sp = ( sup[1] /= (dg = diag[1]) );
rt = ( right[1] /= dg );

for (i=2; i<=n; i++)
  {
  sb = sub[i];
  rt = right[i] = (right[i] - sb * rt) / (dg = diag[i] - sb * sp);
  sp = (sup[i] /= dg);
  }

for (i=n-1; i>=1; i--) right[i] -= sup[i] * right[i+1];
}

//***************************************************************************

void FieldObj::getCurrents(TokenString &Params, int simID, class VarList *VL) 
{
  long pos;
  double res, *ptr;
  int  si, c0, c1;
  
  if (equal(ID,"Ca")) {
     c0 = Params.token_count("current");
     c1 = Params.token_count("currents");
  } else {
     c0 = Params.token_count(ID,"current");
     c1 = Params.token_count(ID,"currents");
  } 

  if (source_num == 0) return;
  if (c0 && c1) globalError(makeMessage("May not mix \"current\" and \"currents\" commands in the same file"));

  FieldObj::SAME_CURRENT = (c0 > 0);

  if ( FieldObj::SAME_CURRENT ) {
     if (equal(ID,"Ca")) pos = Params.token_index("current", simID) + 1;
	 else pos = Params.token_index(ID,"current", simID) + 1;
     if ( Params.equal(pos, VAR_TOKEN) || Params.equal(pos, ASSIGN_TOKEN) ) pos++;
     Current = new ExpressionObj(Params, pos, "Bad expression for the current", VL, 0, &Time, "t");
     Current->Evaluate();
  } else {
     Current = 0;
     if (equal(ID,"Ca")) pos = Params.token_index("currents", simID) + 1;
	 else pos = Params.token_index(ID,"currents", simID) + 1;
     if ( Params.equal(pos, VAR_TOKEN) || Params.equal(pos, ASSIGN_TOKEN) ) pos++;
     for (si = 0; si < source_num; si++) 
       if ( Params.isConst(Params[pos + si], &res) ) { 
         Currents[si].type = NUMBER_TYPE;
         Currents[si].val = res; 
       }
       else if ( ptr = VL->ResolveID(Params[pos + si]) ) {
         Currents[si].type = POINTER_TYPE;
         Currents[si].ptr = ptr;
       }
       else Params.errorMessage(pos + si, makeMessage("Bad specifier for current #%d: expecting a number or time-dependent variable",si+1));
    }  
}


//***************************************************************************

void FieldObj::evaluateCurrents() 
{
  if (!source_num) return;
  ICa = 0;

  for (int i=0; i<source_num; i++)
    ICa += FieldObj::SAME_CURRENT ? Current->Evaluate() : Evaluate(Currents[i]);
}

//***************************************************************************

void FieldObj::printCurrents(FILE *out) 
{
  if (!source_num) return;

  if (FieldObj::SAME_CURRENT && Current )  fprintf(out, "%g ", Current->Evaluate() );
  else {
    fprintf(out,"{ ");
    for (int i=0; i<source_num; i++) if (Currents) fprintf(out, "%g ", Evaluate(Currents[i]) );
    fprintf(out,"}");
  }
  fprintf(out,"\n");
}

//***************************************************************************

void FieldObj::killCurrents() 
{
  if (!source_num) return;

  if (FieldObj::SAME_CURRENT) { if (Current) delete Current; Current = 0; }
}

//****************************************************************************
//*          C L A S S   F I E L D :   I M P L E M E N T A T I O N
//****************************************************************************


FieldObj::FieldObj(TokenString &params, const char *id) : VectorObj(Size)
{
  int i;

  integrate = params.assert("current.shape","square") ? &square : &gaussian;

  source_num = params.token_count(id, "source");

  //params.checkName(id, "concentration field");
  allocate(id);

  Time = 0.0;  
  *this = bgr = 0.0;

  params.get2_param(id, "D", &D, " undefined coefficient of diffusion");   
  if (VERBOSE)
    fprintf(stdlog, "\n### Concentration field %s: diffusion coefficient = %g um^2 / ms\n", ID, D);

  tortDefined = false;
  long tortIndex = 0;                             // this blocks sets tortuosity
  Diff = 0;
  params.token_count(id, "tortuosity", &tortIndex);

  if ( tortIndex ) {
    tortDefined = true;
    Diff = new double[ Size ];
    if (VERBOSE) fprintf(stdlog,"\n#### Tortuosity function = ");
    setFieldByFunction( params, tortIndex + 1, Diff, "Bad tortuosity expression" ); 
    if (VERBOSE) fprintf(stdlog,"\n");
  }

  if (params.token_count(id, "bgr")) {
    bgr = params.get2_param(id, "bgr");
    if (bgr < 0) 
       globalError( makeMessage("Background %s concentration cannot be negative ([%s]bgr = %g)", ID, ID, bgr) );
    *this = bgr;
    if (VERBOSE)  fprintf(stdlog, "    Background [%s] = %g uM \n", ID, bgr);
  }

  long pos;
  if (params.token_count(id, "import", &pos))
    { 
    char *filename = params.get_string(pos + 1);
    if (VERBOSE) fprintf(stdlog, "  # Importing %s from file %s\n", ID, filename);
    import(filename);
    delete filename;
    }

 if (source_num)
  {
  double x, y, z, sigmax, sigmay, sigmaz;
  if (VERBOSE) fprintf(stdlog, "    %s has %d sources:\n", ID, source_num);
  for (i = 1; i <= source_num; i++)
  try {
    x = y = z = 0;
    sigmax = 0.0; sigmay = sigmaz = -1;

    switch (DIMENSIONALITY) {
    case 1:
       params.trail_pars(id, "source", i, 'd', &x, 'd', &sigmax);
       break;
    case 2: 
       params.trail_pars(id, "source", i, 'd', &x, 'd', &y, 'd', &sigmax, 'd', &sigmay);
       break;
    case 3:
       params.trail_pars(id, "source", i, 'd', &x, 'd', &y, 'd', &z, 'd', &sigmax, 
                          'd', &sigmay,  'd', &sigmaz);
       break;
    }

    if (sigmay < 0) sigmay = sigmax;  //  by default widths in all directions are equal
    if (sigmaz < 0) sigmaz = sigmay;  //

    set_source(i-1, x, y, z, sigmax, sigmay, sigmaz); 
    } 
    catch (char *str) { 
      params.errorMessage( params.token_index(id, "source", i) + 1, str, "Cannot initialize this source"); 
    }
  }

 //******************************************************************
  
  extObstNum   = params.token_count(ID, "obstacle");
  extObstArray = extObstNum ? new ObstBoxObj[extObstNum] : 0;
  double x0, x1, y0, y1, z0, z1;

  if (extObstNum) for (i = 0; i < extObstNum; i++)
  try {
    pos = params.token_index(ID, "obstacle", i + 1);
    x0 = x1 = y0 = y1 = z0 = z1 = 0.0;

    if (DIMENSIONALITY == 1)
      params.trail_pars(pos, 'd', &x0, 'd', &x1, 'E');

    else  if (DIMENSIONALITY == 2)
      params.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');

    else  // 3 spatial dimensions
      params.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    

    if (VERBOSE) { fprintf(stdlog,"    Adding special obstacle #%d for field %s:", i+1, id); fflush(stdlog); }

    extObstArray[i].set(x0, x1, y0, y1, z0, z1);

    if (VERBOSE)  {
      fprintf(stdlog," [%g,%g]", x0, x1);
      if (DIMENSIONALITY > 1) fprintf(stdlog, " x [%g,%g]", y0, y1);
      if (DIMENSIONALITY > 2) fprintf(stdlog, " x [%g,%g]", z0, z1);
      fprintf(stdlog, "\n");
    }

    if ( x0 > x1 || y0 > y1 || z0 > z1 )   throw StrCpy("Check diffusion obstacle dimensions"); 
  }
  catch(char *message)
    { params.errorMessage( pos + 1, message); }


  //****************************************************************

  int defs  = params.token_count(id, "bc");
  int boxes = Region->get_box_num() + extObstNum;

  char bc1[20], bc2[20], bc3[20], bc4[20], bc5[20], bc6[20];

  for (i = 1; i<= boxes; i++)
    try { 

    if ( i > defs )   // no b.c. specification for box i:
      {
      if (VERBOSE) fprintf(stdlog,"    Box #%d: not specified; assuming all boundaries are zero flux\n", i);
      set_bc(i-1, "Noflux");     
      }
    else if ( params.equal( params.token_index(id, "bc", i) + 1, "all") ) 
      {
      params.trail_pars(id, "bc", i, 0,0, 's', bc1, 'E');
      if (VERBOSE) fprintf(stdlog,"    Box #%d: all boundaries are %s\n", i, bc1);
      set_bc(i-1, bc1);
      }
    else if (DIMENSIONALITY == 1)
      {
      params.trail_pars(id, "bc", i, 's', bc1, 's', bc2, 'E');
      if (VERBOSE) 
        fprintf(stdlog,"    Box #%d: boundary cond'ns: xmin:%s xmax:%s\n", i, bc1, bc2);
      set_bc(i-1, bc1, bc2, bc2, bc2, bc2, bc2);
      }
    else if (DIMENSIONALITY == 2)
      {
      params.trail_pars(id, "bc", i, 's', bc1, 's', bc2, 's', bc3, 's', bc4,  'E');
      if (VERBOSE) 
        fprintf(stdlog,"    Box #%d: boundary cond'ns: xmin:%s xmax:%s ymin:%s ymax:%s\n", i, bc1, bc2, bc3, bc4);
      set_bc(i-1, bc1, bc2, bc3, bc4, bc4, bc4);
      }
    else  // DIMENSIONALITY == 3
      {
      params.trail_pars(id, "bc", i,
          's', bc1, 's', bc2, 's', bc3, 's', bc4, 's', bc5, 's', bc6, 'E');
      if (VERBOSE) 
        fprintf(stdlog,"    Box #%d: boundary cond'ns: xmin:%s xmax:%s ymin:%s ymax:%s zmin:%s zmax:%s\n", 
                i, bc1, bc2, bc3, bc4, bc5, bc6);
      set_bc(i-1, bc1, bc2, bc3, bc4, bc5, bc6);
      }
    }
  catch(char *str) { params.errorMessage( params.token_index(id, "bc", i) + 1, str); }

  init_boundary();

  // initialAve = bgr;
}

//***************************************************************************

FieldObj::FieldObj(const FieldObj &f) : VectorObj(f)
{
  int nx, ny, nz;
  source_num = f.source_num;
  allocate( f.ID );
 
  if (source_num)
    {
    Current = f.Current;
    for(int i = 0; i < source_num; i++)  
      {
      Currents[i] = f.Currents[i];
      source_x0[i] = f.source_x0[i];
      nx = source_nx[i] = f.source_nx[i];
      source_y0[i] = f.source_y0[i];
      ny = source_ny[i] = f.source_ny[i];
      source_z0[i] = f.source_z0[i];
      nz = source_nz[i] = f.source_nz[i];
      
      source_xsum[i] = new double[nx];
      source_ysum[i] = new double[ny];
      source_zsum[i] = new double[nz];

      int ii;
      for (ii = 0; ii < nx; ii++) source_xsum[i][ii] = f.source_xsum[i][ii];
      for (ii = 0; ii < ny; ii++) source_ysum[i][ii] = f.source_ysum[i][ii];
      for (ii = 0; ii < nz; ii++) source_zsum[i][ii] = f.source_zsum[i][ii];
      }
    }

  extObstArray = 0;

  if (extObstNum = f.extObstNum) {

    extObstArray = new ObstBoxObj[extObstNum]; 
    for (int i=0; i < extObstNum; i++)
      extObstArray[i].set( f.extObstArray[i] );
  }


  for(long i=0; i< size; i++) ptype[i] = f.ptype[i];
  for(int n=0; n< Region->get_surface_num(); n++)  bccond[n] = f.bccond[n];

  D = f.D;
  Time = f.Time;
  bgr = f.bgr;
  Diff = f.Diff;
  // initialAve = f.initialAve;
}

//***************************************************************************

void FieldObj::allocate(const char *id)
{
ID = StrCpy(id);

 if ( !equal(ID,"Ca") ) source_num = 0;

  bccond = new int[Region -> get_surface_num()];

  ptype = new long[size];

  Currents     = new struct TermStruct[source_num];
  source_x0    = new int[source_num];
  source_nx    = new int[source_num];
  source_y0    = new int[source_num];
  source_ny    = new int[source_num];
  source_z0    = new int[source_num];
  source_nz    = new int[source_num];
  source_xsum  = new double *[source_num];
  source_ysum  = new double *[source_num];
  source_zsum  = new double *[source_num];

}

//****************************************************************************

FieldObj &FieldObj::operator=(const FieldObj &f)
{
 *((VectorObj *)this) = f;
 Time = f.Time;
 return *this;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const VectorObj &v)
{
  *((VectorObj *)this) = v;

  //for (long i = 0; i < size; ++i) elem[i] = v[i]; 

return *this;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const double val)
{
*((VectorObj *)this) = val;

//for (long i = 0; i < size; ++i) elem[i] = val; 
return *this;
}

//************************************************************************

FieldObj &FieldObj::operator=(double (*f0)(double, double, double) )
{
long   i = -1;

for (int iz = 0; iz < zsize; iz++)
  for (int iy = 0; iy < ysize; iy++)
    for (int ix = 0; ix < xsize; ix++)
      if ( ptype[++i] & VARY_MASK )
         elem[i] = (*f0)(xcoord[ix], ycoord[iy], zcoord[iz]);
 
return *this;
}


//************************************************************************************
//                   F I E L D   D E S T R U C T O R
//************************************************************************************

  FieldObj::~FieldObj()
    {
      //  fprintf(stderr, "destroying %s\n",ID); fflush(stderr); 

    delete [] ptype;
    delete [] bccond; 

    for (int i = 0; i < source_num; i++)   {
      delete [] source_xsum[i]; delete [] source_ysum[i]; delete [] source_zsum[i];
      }

    if (extObstNum) delete [] extObstArray;

    delete [] source_xsum; delete [] source_ysum; delete [] source_zsum;

    delete [] source_x0;   delete [] source_nx;
    delete [] source_y0;   delete [] source_ny;
    delete [] source_z0;   delete [] source_nz; 

    delete [] ID;

  }


//************************************************************************************
//                   F I E L D   I N I T I A L I Z A T I O N
//************************************************************************************


void  FieldObj::set_bc(int boxid, int bcx0, int bcx1, int bcy0, int bcy1, int bcz0, int bcz1)
{

  if (boxid >= Region->get_box_num() + extObstNum)  
    throw makeMessage("Cannot set boundary condition: box ID=%d > total number of volumes = %d", 
                      boxid, Region->get_box_num());

  if ( (bcx0 > bc_type_num) || (bcx1 > bc_type_num) || (bcy0 > bc_type_num) ||
       (bcy1 > bc_type_num) || (bcz0 > bc_type_num) || (bcz1 > bc_type_num) )  
    throw makeMessage("Bad boundary condition ID for geometry element #%d", boxid);
  
  bccond[boxid*6]   = bcx0;
  bccond[boxid*6+1] = bcx1;
  bccond[boxid*6+2] = bcy0;
  bccond[boxid*6+3] = bcy1;
  bccond[boxid*6+4] = bcz0;
  bccond[boxid*6+5] = bcz1;

}

//******************************************

void  FieldObj::set_bc(int boxid, const char *bcx0, const char *bcx1, const char *bcy0,
                                  const char *bcy1, const char *bcz0, const char *bcz1)
{
  set_bc( boxid, BCarray->bcidtoint(bcx0), BCarray->bcidtoint(bcx1), 
                 BCarray->bcidtoint(bcy0), BCarray->bcidtoint(bcy1), 
	         BCarray->bcidtoint(bcz0), BCarray->bcidtoint(bcz1) );
}

//******************************************

void  FieldObj::set_bc(int boxid, int bc) {
  set_bc( boxid, bc, bc, bc, bc, bc, bc);
}

//******************************************

void  FieldObj::set_bc(int boxid, const char *bcid) {
  set_bc( boxid, BCarray->bcidtoint(bcid) ); 
}

//************************************************************************************
//                              S E T   S O U R C E
//************************************************************************************

void  FieldObj::set_source(int id, double x, double y, double z, double sigmax, double sigmay, double sigmaz)
{

  if (VERBOSE) {
     fprintf(stdlog, "    Setting source #%d at (%g",id+1, x);
     if (DIMENSIONALITY > 1) fprintf(stdlog, ",%g",y);
     if (DIMENSIONALITY > 2) fprintf(stdlog, ",%g",z);
     fprintf(stdlog, "), with half-width of (%g", sigmax);
     if (DIMENSIONALITY > 1) fprintf(stdlog, ",%g",sigmay);
     if (DIMENSIONALITY > 2) fprintf(stdlog, ",%g",sigmaz);
     fprintf(stdlog, ")");
     fprintf(stdlog, "\n");
  }

  if (id >= source_num)
    throw makeMessage("Bad source definition: source id=%d >= source_num=%d",id,source_num);

  if ( !Grid->in_bounds(x, y, z) ) {
    if (DIMENSIONALITY > 2) 
      throw makeMessage("Cannot initialize Ca source: location (%g,%g,%g) out of bounds",x,y,z);
    else if (DIMENSIONALITY > 1) 
      throw makeMessage("Cannot initialize Ca source: location (%g,%g) out of bounds",x,y);
    else  
      throw makeMessage("Cannot initialize Ca source: location (%g) out of bounds",x);
  }

  long ind = Grid->location_to_index(x, y, z);
  
  int ix, iy, iz;
  Grid->split(ind, ix, iy, iz);
  if (VERBOSE) {
     fprintf(stdlog, "    closest node: (%d", ix);
     if (DIMENSIONALITY > 1) fprintf(stdlog, ",%d",iy);
     if (DIMENSIONALITY > 2) fprintf(stdlog, ",%d",iz);
     fprintf(stdlog, ") <--> (%.4g", xcoord[ix]);
     if (DIMENSIONALITY > 1) fprintf(stdlog, ",%.4g",ycoord[iy]);
     if (DIMENSIONALITY > 2) fprintf(stdlog, ",%.4g",zcoord[iz]);
     fprintf(stdlog, ")\n");
  }

  //*************************

  if (VERBOSE > 2) { fprintf(stdlog,"      non-zero current on %s-nodes ",LABEL_DIM1); fflush(stdlog); }
  set_source_xyz(id, x, ind, ix, sigmax, xcoord, xgrid, SURF_XMIN, SURF_XMAX, 1,
                 xsize - 1, source_x0, source_nx, source_xsum);

  if ( DIMENSIONALITY > 1) {
    if (VERBOSE > 2) { fprintf(stdlog,"      non-zero current on %s-nodes ",LABEL_DIM2); fflush(stdlog); }
    set_source_xyz(id, y, ind, iy, sigmay, ycoord, ygrid, SURF_YMIN, SURF_YMAX, xsize,
                 ysize - 1, source_y0, source_ny, source_ysum);
  } else { 
    source_y0[id] = 0; source_ny[id] = 1;
    source_ysum[id] = new double[1]; source_ysum[id][0] = 1.0;
    }

  if ( DIMENSIONALITY > 2) {
    if (VERBOSE > 2)  { fprintf(stdlog,"      non-zero current on %s-nodes ",LABEL_DIM3); fflush(stdlog); }
    set_source_xyz(id, z, ind, iz, sigmaz, zcoord, zgrid, SURF_ZMIN, SURF_ZMAX, xysize,
                 zsize - 1, source_z0, source_nz, source_zsum);
    }
  else { 
    source_z0[id] = 0; source_nz[id] = 1;
    source_zsum[id] = new double[1]; source_zsum[id][0] = 1.0;
    }
}


//************************************************************************************
//                            S O U R C E   S P R E A D
//************************************************************************************

void  FieldObj::set_source_xyz(int id, double c0, long i0, int ix, double sigma, 
               double *coord, double *grid, int minMask, int maxMask, int incr, int N,
               int *source_coord0, int *interval_num, double **source_weights )
{
if (sigma < 0.1 * grid[ix])  sigma = 0.1 * grid[ix];  // if source is point-like, set half-width at 
                                                      // 1/10th of grid spacing
  double *temp_weights = new double[N+1];
  double check_sum = 0.0;

  bool flag = false;    // to flag the entry into the region of non-zero current

  long i3d = i0;
  int  ileft = ix, iright = ix;
  while ( !(point_type(i3d) & minMask) ) { i3d -= incr; ileft --; } 
  i3d = i0;
  while ( !(point_type(i3d) & maxMask) ) { i3d += incr; iright ++; }

  for (int ii = ileft; ii <= iright; ii++) 
    {
    double dx0 = 0.5 * ( (ii > ileft)  ? grid[ii-1] : 0.0 );
    double dx1 = 0.5 * ( (ii < iright) ? grid[ii] : 0.0 );

    double x0 = coord[ii] - c0 - dx0;
    double x1 = coord[ii] - c0 + dx1;
    
    double left0 = 2 * (c0 - coord[ileft]) + x0;
    double left1 = 2 * (c0 - coord[ileft]) + x1;

    double right0 = 2 * (coord[iright] - c0) - x1;
    double right1 = 2 * (coord[iright] - c0) - x0;

    temp_weights[ii] = integrate( x0, x1, sigma, 0 ) + 
                       integrate( left0, left1, sigma, 0 ) + 
                       integrate( right0, right1, sigma, 0 ); 
    double dv = (dx0 + dx1);  // default volume element

    if (coord == xcoord) {   // check if geom1 == r or geom1 == rho
      if (GEOMETRY1 == 1) {
         if (c0 == 0.0) temp_weights[ii] = integrate( x0, x1, sigma, 1 );  // if geom2 == phi, no pi multiplier:
         dv = (pow( coord[ii] + dx1, 2 ) - pow( coord[ii] - dx0, 2 )) * ( (GEOMETRY2 == 2) ? 0.5 : pi ); 
      }
      else if (GEOMETRY1 == 2) {
         if (c0 == 0.0) temp_weights[ii] = integrate( x0, x1, sigma, 2 );
         dv = (pow( coord[ii] + dx1, 3 ) - pow( coord[ii] - dx0, 3 )) / 3.0;
         if (GEOMETRY == SPHERICAL) dv *= (4 * pi); else if (GEOMETRY == CONICAL) dv *= (2 * pi);
      }
    } else if (coord == ycoord) {   // check if geom2 == theta
      if (GEOMETRY2 == 3) {
         if (c0 == 0.0) temp_weights[ii] = integrate( x0, x1, sigma, 1 ); 
         dv = cos( coord[ii] - dx0 ) - cos( coord[ii] + dx1 );
      }
    }
  
   // else dv has the default value of (dx0 + dx1)
    
    check_sum += temp_weights[ii];
    temp_weights[ii] /= dv;
   
    if ( temp_weights[ii] > SOURCE_EPS )   // integral between ii-1/2 and ii+1/2 > eps
      {
       if ( !flag )  
	 {
         flag = true;
         source_coord0[id] = ii;
         interval_num[id] = 1;
	 }
       else interval_num[id]++;
       if (VERBOSE > 2) fprintf(stdlog, "%d(%.5g):%.4g ", ii, coord[ii], temp_weights[ii]*dv);
      }
    } // end for (ii = ileft; ii <= iright; ii++) 

  //*************************

  source_weights[id] = new double[interval_num[id]];

  for (int i = 0; i < interval_num[id]; i++) 
    source_weights[id][i] = temp_weights[i + source_coord0[id]]; // / check_sum;   // renormalize 

  if (fabs(check_sum - 1) > 1.0e-4) fprintf(stdlog,"*** WARNING: current total norm error=%.2e\n", check_sum - 1.0);
  // fprintf(stdlog,"*** WARNING: current total norm error=%.2e\n", check_sum - 1.0);

  if (VERBOSE > 2) fprintf(stdlog, "\n");

  delete [] temp_weights;
}

//************************************************************************************
//                      O L D    S O U R C E   S P R E A D
//************************************************************************************
/*  
 
  Evaluating the integral of exp(-x^2/sigma^2) x^dim dx
  Normalization factor is evaluated by setting to 1 the intergal from 0 to inf:

  Int( exp(-x^2/sigma^2) x^dim dx, 0, inf ) = sigma^(dim+1) Int( exp(-y*y) y^dim dy, 0, inf) =
  = sigma^(dim+1) / 2 * Int( exp(-y*y) y^(dim-1) d(y^2), 0 , inf) =
  = sigma^(dim+1) / 2 * Gamma( (dim+1) / 2)
  
  So the normalization factor is 2 / (sigma^(dim+1) Gamma( (dim+1) / 2) ).
  For dim=0, integration is from -inf to inf, not from 0 to inf, so the factor is half that

  dim = 0: 1/(sigma sqrt(pi))
  dim = 1: 2/(sigma^2);
  dim = 2: 4/(sigma^3 sqrt(pi)) 

************************************************************************************/

double gaussian(double x0, double x1, double sigma, int dim)
{
  double eps = 1.0e-10;

  const double a[3] = { 1.0 / ( sigma * sqrt(pi) ),
                        2.0 / ( sigma * sigma ),
			4.0 / ( sigma * sigma * sigma * sqrt(pi) )};
  double b = -1.0 / (sigma * sigma);

  long   N = 2;
  double resN = 0, res2N = 0, res4N = 0;
  double x, dx, sum;

  do
    {
    resN = res2N;
    N *= 2;
    dx = (x1 - x0) / double(N);
    sum = 0.0;
    for (int n = 1; n < N; n++)
      {
      x = x0 + dx * double(n);
      sum += ((n & 1) + 1) * exp( b * x * x ) * pow(x, dim);
      }
    res2N = dx / 3.0 * a[dim] * ( 2 * sum + exp( b * x0 * x0 ) * pow(x0, dim) + exp( b * x1 * x1 ) * pow(x1, dim) );
    res4N = (16.0 * res2N - resN) / 15.0;
    }
  while ( fabs(res4N - res2N) > eps * fabs(res4N + res2N) && fabs(res4N + res2N) > 1e-12);

  return res4N;
}

//************************************************************************************/

double square(double x0, double x1, double sigma, int dim)
{
  if (x0 >= sigma || x1 <= -sigma) return 0;

  double norm = 1.0 / pow(sigma, double(dim + 1) );
  if (dim == 0) norm /= 2.0;

  if (x0 >= 0) {
     if (x1 < sigma) return  norm * (pow(x1,    dim + 1) - pow(x0, dim + 1) );
                else return  norm * (pow(sigma, dim + 1) - pow(x0, dim + 1) );
  } else {  // must be 1D (dim=0) case; already know that x1 > -sigma
     if (x0 < -sigma) if (x1 < sigma) return norm * (x1 + sigma);
                      else return 1.0;
                 else if (x1 < sigma) return norm * (x1 - x0);
                      else return norm * (sigma - x0);
  }
 
}


//***************************************************************

void  FieldObj::add_sources(VectorObj &f, double dt)
{
int    i, ix, iy, iz;
long   xind, yind, zind;
double addx, addy, charge, ica;

  ICa = 0.0; 

  for (i=0; i < source_num; i++) {

    ica = SAME_CURRENT ? Current->Evaluate() : Evaluate(Currents[i]);
    if (ica < MIN_CURRENT) continue;
    charge = ica * dt;
    ICa += ica;
    int nx = source_nx[i];
    int ny = source_ny[i];
    int nz = source_nz[i];
    xind = source_x0[i] + xsize * source_y0[i] + xysize * source_z0[i];
    //    fprintf(stderr,"%d %d %d  %d %d %d\n\n",nx,ny,nz, source_x0[i],source_y0[i],source_z0[i]);

    for (ix = 0; ix < nx; ix++, xind ++) {
      addx = charge * source_xsum[i][ix];
      for (iy = 0, yind = xind; iy < ny; iy++, yind += xsize) {
	addy = addx * source_ysum[i][iy];
        for (iz = 0, zind = yind; iz < nz; iz++, zind += xysize) //{ 
           f.elem[zind] += addy * source_zsum[i][iz];
	   //fprintf(stderr,"adding to f[%ld]=%g+%g\n",zind, f.elem[zind], addy * source_zsum[i][iz]); }
      }
    }
  }

}

//************************************************************************
//            B O U N D A R Y   I N I T I A L I Z A T I O N
//************************************************************************
/*
   RegionObj::point_type(int,int,int) bit fields are:
   bit # is:      ...8       |   7    |   6  |  5   |  4   |  3   |  2   |  1  
   meaning:    box id number | inside | zmax | zmin | ymax | ymin | xmax | xmin

  ***************************************************************

   ptype array bit fields are:

   bit # is:  ...10    |    9    |    8      |   7    |   6  |  5   |  4   |  3   |  2   |  1  
   meaning:  bc_type # | outside | dirichlet | inside | zmax | zmin | ymax | ymin | xmax | xmin
  
   bit # is: |   30...24    |   23...17    |   16...10    |
   meaning:  | z bc_type id | y bc_type id | x bc_type id |

  ***************************************************************/

void FieldObj::init_boundary()
{
long      i;
int       p, bc = 0, boxid;
int       ix, iy, iz;

 for (i = 0; i < size; i++) ptype[i] =  _INSIDE_;   // default: point inside

 for (i = 0; i < size; i++)
   {
   Grid->split(i, ix, iy, iz);
   p = point_type(ix, iy, iz); 
   boxid = p >> 7;
   p = p & INSIDE_MASK;              // remove box id (keep bits 1 through 7)
   if (p == 0)                       // point is outside
     {
     ptype[i] = _OUTSIDE_; 
     elem[i] = 0.0;
     }
   else if (p & SURF_MASK)           // point on the surface
     {
     if (p & SURF_XMIN)
       if (bc_deriv[bc = get_bc(boxid,0)] != 0) ptype[i] |= (bc << XSHIFT) + SURF_XMIN;
       else  { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_XMAX) 
       if (bc_deriv[bc = get_bc(boxid,1)] != 0) ptype[i] |= (bc << XSHIFT) + SURF_XMAX;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; } 

     if (p & SURF_YMIN) 
       if (bc_deriv[bc = get_bc(boxid,2)] != 0) ptype[i] |= (bc << YSHIFT) + SURF_YMIN;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_YMAX) 
       if (bc_deriv[bc = get_bc(boxid,3)] != 0) ptype[i] |= (bc << YSHIFT) + SURF_YMAX;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }

     if (p & SURF_ZMIN)
       if (bc_deriv[bc = get_bc(boxid,4)] != 0) ptype[i] |= (bc << ZSHIFT) + SURF_ZMIN;
       else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_ZMAX)
       if (bc_deriv[bc = get_bc(boxid,5)] != 0) ptype[i] |= (bc << ZSHIFT) + SURF_ZMAX;
       else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }

     } // end if (p & SURF_MASK)
  
  } // end for loop


 for (i = 0; i < size; i++)   // SECOND PASS: take care of Dirichlet bc
   if ( ptype[i] & DIRICHLET ) 
     {
     elem[i] = bc_coef[ bc = ptype[i] >> XSHIFT ];
     Grid->split(i, ix, iy, iz);
     p = point_type(ix, iy, iz); 
 
     if ( ix + 1 < xsize ) if ((point_type(ix+1,iy,iz) & _INSIDE_) && (ptype[i+1] & VARY_MASK))
       ptype[i + 1] |= (bc << XSHIFT) + SURF_XMIN; 

     if ( ix - 1 > 0 )     if ((point_type(ix-1,iy,iz) &  _INSIDE_) && (ptype[i-1] & VARY_MASK))
       ptype[i - 1] |= (bc << XSHIFT) + SURF_XMAX; 

     if ( iy + 1 < ysize ) if ((point_type(ix,iy+1,iz) & _INSIDE_) && (ptype[i+xsize] & VARY_MASK))
       ptype[i + xsize] |= (bc << YSHIFT) + SURF_YMIN; 
     
     if ( iy - 1 > 0 )   if ((point_type(ix,iy-1,iz) & _INSIDE_) && (ptype[i-xsize] & VARY_MASK))
       ptype[i - xsize] |= (bc << YSHIFT) + SURF_YMAX; 

     if ( iz + 1 < zsize ) if ((point_type(ix,iy,iz+1) & _INSIDE_) && (ptype[i+xysize] & VARY_MASK))
        ptype[i + xysize] |= (bc << ZSHIFT) + SURF_ZMIN; 
     
     if ( iz - 1 > 0 ) if ((point_type(ix,iy,iz-1) & _INSIDE_) && (ptype[i-xysize] & VARY_MASK))
        ptype[i - xysize] |= (bc << ZSHIFT) + SURF_ZMAX; 

  } // end for loop

}


//*****************************************************************************
/*
double FieldObj::interpolate(double x, double y, double z)
{
int ix = 0, iy = 0, iz = 0;

while (x > xcoord[ix] && ix < (xsize - 1) ) ix++;
while (y > ycoord[iy] && iy < (ysize - 1) ) iy++;
while (z > zcoord[iz] && iz < (zsize - 1) ) iz++;

long i = ix + xsize * iy + xysize * iz;

double p1 = (x - xcoord[ix-1]) / (xcoord[ix] - xcoord[ix-1]);
double p2 = (y - ycoord[iy-1]) / (ycoord[iy] - ycoord[iy-1]);
double p3 = (z - zcoord[iz-1]) / (zcoord[iz] - zcoord[iz-1]);

double res =      p1  *      p2  *      p3  * elem[i                     ] 
           + (1 - p1) * (1 - p2) * (1 - p3) * elem[i - 1 - xsize - xysize]
           +      p1  * (1 - p2) * (1 - p3) * elem[i     - xsize - xysize]
           + (1 - p1) *      p2  * (1 - p3) * elem[i - 1         - xysize]
           + (1 - p1) * (1 - p2) *      p3  * elem[i - 1 - xsize         ]
           +      p1  *      p2  * (1 - p3) * elem[i             - xysize]
           +      p1  * (1 - p2) *      p3  * elem[i     - xsize         ]
           + (1 - p1) *      p2  *      p3  * elem[i - 1                 ];

return res;
}
*/
//***********************************************************************************
/*
double FieldObj::get_value(double x, double y, double z)
{
return elem[location_to_index(x, y, z)];
}
*/
//***********************************************************************************

double FieldObj::sum()
{
double s = 0.0;
for(long i=0; i<size; i++) if (ptype[i] & REGION_MASK) s += elem[i];
return s;
}

//***********************************************************************************

double FieldObj::averageSpherical(double *V)
{
 double s = 0.0, v = 0.0, dv;

 for (int i = 1; i < xsize; i++ )

   if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) )

     if ( xcoord[i-1] == 0.0 ) {                            // int( f(r) r^2, r=0..dr ) = 
       v += ( dv = 2 / 3.0 * pow( xcoord[i], 3 ) );         // int( ( f0 + r/dr (f1-f0) ) r^2, r=0..dr) =
       s += 3.0 / 4.0 * dv * (elem[i] + elem[i-1] / 3.0 );  // dr^3 ( f0/3 + (f1-f0)/4 ) = dr^3 / 4 ( f0/3 + f1 )
     }        
     else  {
       v += ( dv = xgrid[i-1] * (xcoord[i-1] * xcoord[i-1] + xcoord[i] * xcoord[i] ) );
       s += xgrid[i-1] * (elem[i-1] * xcoord[i-1] * xcoord[i-1] + elem[i] * xcoord[i] * xcoord[i]);
     }

if (V) *V = 2 * pi * v;
return s / v ;
}

//***********************************************************************************
/*
double FieldObj::averageConical(double *V)  // calculate integral( field[r,theta] sin(theta) r^2 dr dtheta )
{
 double s = 0.0, v = 0.0, Ith0, Ith1, Ir0, Ir1, rp, rm, thp, thm, dyInv, dvInv;

 for (int iy = 1; iy < ysize; iy++ )  {

   Ith0 = cos( ycoord[iy-1] ) - cos( ycoord[iy] );             // integral of sin(theta) dtheta
   Ith1 = sin( ycoord[iy] ) - sin( ycoord[iy-1] ) +            // integral of theta sin(theta) dtheta
          ycoord[iy-1] * Ith0 - ygrid[iy-1] * cos( ycoord[iy] );
   thp = Ith1 - ycoord[iy-1] * Ith0;
   thm = ycoord[iy] * Ith0 - Ith1;
   dyInv = 1 / ygrid[iy-1];

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       Ir0 = ( pow( xcoord[ix], 3) - pow( xcoord[ix-1], 3) ) / 3.0;  // integral of r^2 dr
       Ir1 = ( pow( xcoord[ix], 4) - pow( xcoord[ix-1], 4) ) / 4.0;  // integral of r r^2 dr
       rp = Ir1 - xcoord[ix-1] * Ir0;
       rm = xcoord[ix] * Ir0 - Ir1;
       dvInv = dyInv / xgrid[ix-1];

       s += dvInv * ( elem[i-1-xsize] * rm * thm + elem[i-1] * rm * thp + 
                                          elem[i-xsize] * rp * thm + elem[i] * rp * thp );
       v += dvInv * ( rm + rp ) * ( thm + thp );
     }

   }
 }

if (V) *V = v;
return s / v ;
}
*/
//***********************************************************************************
 
double FieldObj::averageConical(double *V)
{
 double s = 0.0, v = 0.0, dv, dvy;

 for (int iy = 1; iy < ysize; iy++ )  {

   dvy = cos(ycoord[iy-1]) - cos(ycoord[iy]);

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       v += ( dv = dvy * ( pow(xcoord[ix],3) - pow(xcoord[ix-1],3) ) );
       s += dv * (elem[i-1-xsize]  + elem[i-1] + elem[i-xsize] + elem[i]);
     }
   }
 }

if (V) *V = 2 * pi / 3.0 * v;
return 0.25 * s / v;
}

//***********************************************************************************

double FieldObj::averageSpherical3D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = pow(xcoord[ix],3) - pow(xcoord[ix-1],3);
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = dvx * ( cos(ycoord[iy-1]) - cos(ycoord[iy]) );
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 if (V) *V = v / 3.0;
 return s / v * 0.125;
}

/*
// Average over nodes, as opposed to an average over cells (as in above) 

double FieldObj::averageSpherical3D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 0; ix < xsize; ix++)
  {
  for(iy = 0; iy < ysize; iy++) 
    {
    for(iz = 0; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
      double dy0 = (iy == 0) ? 0 : ( (ptype[i-xsize] & _OUTSIDE_) ? 0: 0.5 * ygrid[iy-1]);
      double dy1 = (iy == ysize - 1) ? 0 : ( (ptype[i+xsize] & _OUTSIDE_) ? 0 : 0.5 * ygrid[iy]);
      dvxy = dvx * ( cos(ycoord[iy]-dy0) - cos(ycoord[iy]+dy1) );
      double dx0 = (ix == 0) ? 0 : ( (ptype[i-1] & _OUTSIDE_) ? 0: 0.5 * xgrid[ix-1]);
      double dx1 = (ix == xsize - 1) ? 0 : ( (ptype[i+1] & _OUTSIDE_) ? 0 : 0.5 * xgrid[ix]);
      dvx = pow(xcoord[ix]+dx1,3) - pow(xcoord[ix]-dx0,3);
      double dz0 = (iz == 0) ? 0 : ( (ptype[i-xysize] & _OUTSIDE_) ? 0: 0.5 * zgrid[iz-1]);
      double dz1 = (iz == zsize - 1) ? 0 : ( (ptype[i+xysize] & _OUTSIDE_) ? 0 : 0.5 * zgrid[iz]);
  
      if ( !(ptype[i] & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * (dz0 + dz1) );
          s += dv * *e;
	}
      }
    }
  }
 if (V) *V = v / 3.0;
 return s / v;
}
*/

//***********************************************************************************

double FieldObj::averageCylindrical3D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = pow(xcoord[ix],2) - pow(xcoord[ix-1],2);
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = ( dvx * ygrid[iy-1] );
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 if (V) *V = 0.5 * v;
 return s / v * 0.125;
}
//***********************************************************************************
 
double FieldObj::averageCylindrical(double *V)
{
 double s = 0.0, v = 0.0, dv, dvy;

 for (int iy = 1; iy < ysize; iy++ )  {

   dvy = ygrid[iy-1];

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       v += ( dv = dvy * ( xcoord[ix] * xcoord[ix] - xcoord[ix-1] * xcoord[ix-1] ) );
       s += dv * (elem[i-1-xsize]  + elem[i-1] + elem[i-xsize] + elem[i]);
     }
   }
 }

if (V) *V = pi * v;
return 0.25 * s / v;
}
//***********************************************************************************

double FieldObj::averagePolar(double *V)
{
 double s = 0.0, v = 0.0, dv, dvy;

 for (int iy = 1; iy < ysize; iy++ )  {

   dvy = ygrid[iy-1];

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       v += ( dv = dvy * ( xcoord[ix] * xcoord[ix] - xcoord[ix-1] * xcoord[ix-1] ) );
       s += dv * (elem[i-1-xsize]  + elem[i-1] + elem[i-xsize] + elem[i]);
     }
   }
 }

 if (V) *V = 0.5 * v;
 return s / v * 0.25;
}
//***********************************************************************************
 
double FieldObj::averageDisc(double *V)
{
 double s = 0.0, v = 0.0, dv;

   for (int i = 1; i < xsize; i++ ) 
     if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) ) {
       v += ( dv = xcoord[i] * xcoord[i] - xcoord[i-1] * xcoord[i-1] );
       s += dv * (elem[i-1] + elem[i]);
     }

if (V) *V = pi * v;
return 0.5 * s / v;
}
//***********************************************************************************

double FieldObj::average3D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = xgrid[ix-1];
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = dvx * ygrid[iy-1];
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 if (V) *V = v;
 return s / v * 0.125;
}


//***********************************************************************************

double FieldObj::average2D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx;
int    ix, iy;
long   i = 0;
double *e;

  for(ix = 1; ix < xsize; ix++) 
    {
    dvx = xgrid[ix-1];
    for(iy = 1; iy < ysize; iy++) 
      {
      e = elem + (i = ix + xsize * iy);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvx * ygrid[iy-1] );
          s += dv * (*e + *(e-1) + *(e-xsize) + *(e-1-xsize) );  
	}
      }
    }
  
 if (V) *V = v;
 return s / v * 0.25;
}

//***********************************************************************************

double FieldObj::average1D(double *V)
{
double s = 0.0, v = 0.0, dv;
long   i = 0;

for(i = 1; i < xsize; i++)
  if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) ) {
    v += (dv = xgrid[i-1]);
    s += dv * (elem[i] + elem[i-1]);
  }

 if (V) *V = v;
 return s / v * 0.5;
}

//***********************************************************************************
/*
double FieldObj::FluxBox(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = xgrid[ix-1];
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = dvx * ygrid[iy-1];
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
      int p = ptype[i];
      if ( p & SURF_MASK ) {

        if ( ( p & SURF_XMIN ) 



      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 if (V) *V = v;
 return s / v * 0.125;
}
*/
//***********************************************************************************

double FieldObj::L2norm()
{
double s = 0.0;
for(long i=0; i<size; i++) if (ptype[i] & REGION_MASK) s += elem[i] * elem[i];
return s / double(size);
}


//***********************************************************************************
/*
void print3d(long *var)
{
for (int iz=0; iz < zsize; iz++, fprintf(stdlog,("\n"))
 for (int iy=0; iy < ysize; iy++, fprintf(stdlog,("\n"))
  for (int ix=0; ix < xsize; ix++) 
      fprintf(stdlog,("%ld ",var[ix + iy * xsize + iz * xysize]);
}
*/  
//***********************************************************************************
/*
void FieldObj::print3d(int n)
{
        char s1[20], s2[20];
        strcpy(s1,"%.");
        sprintf(s2,"%d",n);
        strcat(s1,s2);
        strcat(s1,"f ");

for (int iz=0; iz<zsize; iz++, fprintf(stdlog,"\n"))
 for (int iy=0; iy<ysize; iy++, fprintf(stdlog,"\n"))
  for (int ix=0; ix<xsize; ix++) fprintf(stdlog,s1,elem[ix+iy*xsize+iz*xysize]);

}
*/
//***********************************************************************************

void FieldObj::print_ptype()
{
for (int iz=0; iz < zsize; iz++, fprintf(stdlog,"\n"))
 for (int iy=0; iy < ysize; iy++, fprintf(stdlog,"\n"))
  for (int ix=0; ix < xsize; ix++) 
      print_ptype(ix + iy * xsize + iz * xysize);
}

//***************************************************************************************
//                              X   D I R E C T I O N
//***************************************************************************************
/*
 Find vector Result_new (to be stored in array Result) satisfying

 (1 + c * Dx^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result 

Spherical:
                                                        1                         1
 (1 + c * Dr^2 - Lin) Result_new = ( cx * Dr^2 + cth * --- * Dth^2 + cphi * ------------- * Dphi^2 ) This + Result 
                                                       r^2                  sin(th)^2 r^2
Cylindrical:
                                                         1                  
 (1 + c * Dr^2 - Lin) Result_new = ( cx * Dr^2 + cphi * --- * Dphi^2 + cz * Dz^2 ) This + Result 
                                                        r^2                 

 | diag[1] sup[1]   0      0 :                    |                | right[1] |
 | sub[2]  diag[2] sup[2]  0 :            0       |                |          |
 |...........................:....................| x Result_new = |          |
 |                           : diag[n-1] sup[n-1] |                |          |
 |            0              : sub[n]    diag[n]  |                | right[n] |

 Operator Dx^2:   Dx^2 f[i] == xminus[i] f[ix-1] + xcenter[i] f[i] + xplus[i] f[ix+1]
                          i == (ix,iy,iz)

***************************************************************************************/

void FieldObj::Run3Dx(double c, double cx, double cy0, double cz0, double *Lin, double *Result)
{
int      ix = 0, iy = 0, iz = 0, ix0 = -1, iy0 = -1;
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   xminus, xplus, xcenter, xcoef;
double   yminus, yplus, ycenter, ycoef;
double   zminus, zplus, zcenter, zcoef;
double   ri, cy=cy0, cz=cz0, a=1.0, b=1.0;


if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)  if (ptype[i] & VARY_MASK) 
      Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) );
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK )  {
    n++;
    if ( bcvar & SURF_XMIN ) istart = i; 
    (this->*nablaX)(i, ix, bcvar, xminus, xcenter, xplus, xcoef);
    (this->*nablaY)(i, iy, bcvar, yminus, ycenter, yplus, ycoef);
    (this->*nablaZ)(i, iz, bcvar, zminus, zcenter, zplus, zcoef);

    if ( (GEOMETRY == SPHERICAL3D || GEOMETRY == CYLINDRICAL3D) && (ix0 != ix) ) {
      if (xcoord[ix] == 0) a = 100 / ( xgrid[ix] * xgrid[ix] ); // no nabla theta or nabla phi at origin
      else a = 1 / ( xcoord[ix] * xcoord[ix] ); 
      ix0 = ix; cy = cy0 * a; cz = cz0 * a * b;
    }
    if ( GEOMETRY == SPHERICAL3D && iy0 != iy) {
      if (ycoord[iy] == 0 || ycoord[iy] == pi) b = 10 / sin(ygrid[iy]);   // no nabla phi at theta = 0
      else b = 1 / sin(ycoord[iy]); 
      b *= b; cz = cz0 * a * b;
      iy0 = iy;
    }
    ri = Result[i] + (cx - c) * xcoef + cy * ycoef + cz * zcoef + (cx * xcenter + cy * ycenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cx * xminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cx * xplus  * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cy * yminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cy * yplus  * elem[i+xsize];    
    if ( ! (bcvar & SURF_ZMIN) )  ri += cz * zminus * elem[i-xysize]; 
    if ( ! (bcvar & SURF_ZMAX) )  ri += cz * zplus  * elem[i+xysize];    
 
    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * xcenter;
    sup[n] = c * xplus;  sub[n] = c * xminus;
    right[n] = ri; 

    i ++;
    if (++ix == xsize) break;
  } // while bcvar > 0 and ix < xsize

  if (n)  { 
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
  }

  if (ix < xsize) 
    while ( !(ptype[i] & VARY_MASK) ) {
      if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
      i ++; 
      if ( (++ix) == xsize ) break; 
    }

  if (ix == xsize) { 
     ix = 0;
     if (++iy == ysize) { iy = 0; iz ++; } 
  }
  
  } while (iz < zsize);   
}

//***************************************************************************************
//                               Y   D I R E C T I O N
//
//  (1 + c * Dy^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result
/*
Spherical:
           1                                                 1                         1
 (1 + c * --- Dth^2 - Lin) Result_new = ( cx * Dr^2 + cth * --- * Dth^2 + cphi * ------------- * Dphi^2 ) This + Result 
          r^2                                               r^2                  sin(th)^2 r^2
Cylindrical:
           1                                                   1                  
 (1 + c * --- Dphi^2 - Lin) Result_new = ( cx * Dr^2 + cphi * --- * Dphi^2 + cz * Dz^2 ) This + Result 
          r^2                                                 r^2                 
*/
//***************************************************************************************

void FieldObj::Run3Dy(double c0, double cx, double cy0, double cz0, double *Lin, double *Result)
{
int      ix = 0, iz = 0, iy = 0, ix0 = -1, iy0 = -1;
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   xminus, xplus, xcenter, xcoef;
double   yminus, yplus, ycenter, ycoef;
double   zminus, zplus, zcenter, zcoef;
double   ri, a=1.0, b=1.0, c=c0, cy=cy0, cz=cz0;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++) if (ptype[i] & VARY_MASK) 
     Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) );
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {
    n++;
    if ( bcvar & SURF_YMIN ) istart = i; 
    (this->*nablaX)(i, ix, bcvar, xminus, xcenter, xplus, xcoef);
    (this->*nablaY)(i, iy, bcvar, yminus, ycenter, yplus, ycoef);
    (this->*nablaZ)(i, iz, bcvar, zminus, zcenter, zplus, zcoef);
    
    if ( (GEOMETRY == SPHERICAL3D || GEOMETRY == CYLINDRICAL3D) && (ix0 != ix) ) {
      if (xcoord[ix] == 0) a = 100 / ( xgrid[ix] * xgrid[ix] ); // no nabla theta or nabla phi at origin
      else a = 1 / ( xcoord[ix] * xcoord[ix] );
      cy = cy0 * a; c = c0 * a; cz = cz0 * a * b;
      ix0 = ix;
    }
    if ( GEOMETRY == SPHERICAL3D && iy0 != iy) {
      if (ycoord[iy] == 0 || ycoord[iy] == pi) b = 10 / sin(ygrid[iy]);   // no nabla phi at theta = 0
      else b = 1 / sin(ycoord[iy]); 
      b *= b;
      cz = cz0 * a * b; 
      iy0 = iy;
    }
    ri = Result[i] + (cy - c) * ycoef + cx * xcoef + cz * zcoef + (cx * xcenter + cy * ycenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cx * xminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cx * xplus * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cy * yminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cy * yplus * elem[i+xsize];    
    if ( ! (bcvar & SURF_ZMIN) )  ri += cz * zminus * elem[i-xysize]; 
    if ( ! (bcvar & SURF_ZMAX) )  ri += cz * zplus * elem[i+xysize];    

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * ycenter;
    sup[n] = c * yplus;  sub[n] = c * yminus;
    right[n] = ri; 

    i += xsize;  
    if (++iy == ysize) break;
  } // while bcvar > 0 and iy < ysize

  if (n > 0)  {
      tridiag(n);
      for (int ii = 1, ind = istart; ii<=n; ii++, ind += xsize) Result[ind] = right[ii]; 
      n = 0;
  }

  if (iy < ysize)
      while ( !(ptype[i] & VARY_MASK) ) {  
         if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
         i += xsize; if ( (++iy) == ysize ) break; 
  }

  if (iy == ysize)    {
      iy = 0 ;
      if (++iz == zsize) { i -= size - 1; iz = 0; ix++; }
  }

} while (ix < xsize); 

}


//***************************************************************************************
//                              Z   D I R E C T I O N
//***************************************************************************************
//
// (1 + c * Dz^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result 
/*
Spherical:
                1                                                     1                         1
 (1 + c * ------------ Dth^2 - Lin) Result_new = ( cx * Dr^2 + cth * --- * Dth^2 + cphi * ------------- * Dphi^2 ) This + Result 
          sin(th)^2 r^2                                              r^2                  sin(th)^2 r^2
Cylindrical:
                                                         1                  
 (1 + c * Dz^2 - Lin) Result_new = ( cx * Dr^2 + cphi * --- * Dphi^2 + cz * Dz^2 ) This + Result 
                                                        r^2                 
*/
//***************************************************************************************

void FieldObj::Run3Dz(double c0, double cx, double cy0, double cz0, double *Lin, double *Result)

{
int      iz = 0, ix = 0, iy = 0, ix0 = -1, iy0 = -1;
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   xminus, xplus, xcenter, xcoef;
double   yminus, yplus, ycenter, ycoef;
double   zminus, zplus, zcenter, zcoef;
double   ri, a=1.0, b=1.0, c=c0, cy=cy0, cz=cz0;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++) if (ptype[i] & VARY_MASK) 
    Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) );
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {  
    n++;

    if ( bcvar & SURF_ZMIN ) istart = i; 
    (this->*nablaX)(i, ix, bcvar, xminus, xcenter, xplus, xcoef);
    (this->*nablaY)(i, iy, bcvar, yminus, ycenter, yplus, ycoef);
    (this->*nablaZ)(i, iz, bcvar, zminus, zcenter, zplus, zcoef);
    
    if ( (GEOMETRY == SPHERICAL3D || GEOMETRY == CYLINDRICAL3D) && (ix0 != ix) ) {
      if (xcoord[ix] == 0) a = 100 / (xgrid[ix] * xgrid[ix]);   // no nabla theta or nabla phi at origin
      else a = 1 / (xcoord[ix] * xcoord[ix]); 
      ix0 = ix; cy = cy0 * a; c = c0 * a * b; cz = cz0 * a * b;
    }
    if ( GEOMETRY == SPHERICAL3D && iy0 != iy) {
      if (ycoord[iy] == 0 || ycoord[iy] == pi) b = 10 / sin(ygrid[iy]);   // no nabla phi at theta = 0
      else b = 1 / sin(ycoord[iy]); 
      b *= b;
      iy0 = iy; c = c0 * a * b; cz = cz0 * a * b;
    }
    ri = Result[i] + (cz - c) * zcoef + cx * xcoef + cy * ycoef + (cx * xcenter + cy * ycenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cx * xminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cx * xplus * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cy * yminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cy * yplus * elem[i+xsize];    
    if ( ! (bcvar & SURF_ZMIN) )  ri += cz * zminus * elem[i-xysize]; 
    if ( ! (bcvar & SURF_ZMAX) )  ri += cz * zplus * elem[i+xysize];    

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * zcenter;
    sup[n] = c * zplus;  sub[n] = c * zminus;
    right[n] = ri; 

    i += xysize; ++iz;
    if (iz == zsize) break;

  } // while ptype > 0 and iz < zsize

  if (n > 0)    {  
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind += xysize) Result[ind] = right[ii]; 
    n = 0;
  } 
 
  if (iz < zsize)
    while ( !(ptype[i] & VARY_MASK) ) {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       i += xysize; if ( (++iz) == zsize) break; 
    }

  if (iz == zsize)  {
      iz = 0; i -= (size - 1);
      if (++ix == xsize) { ix = 0; iy ++; }
  }

} while (iy < ysize);

}

/***************************************************************************************
                                    1 D    R U N
                                    ^^^^^^^^^^^^

 Find vector Result_new (to be stored in array Result) satisfying

 (1 + c * Dx^2 - Lin) Result_new = cc * Dx^2 This + Result 

 | diag[1] sup[1]   0      0 :                    |                | right[1] |
 | sub[2]  diag[2] sup[2]  0 :            0       |                |          |
 |...........................:....................| x Result_new = |          |
 |                           : diag[n-1] sup[n-1] |                |          |
 |            0              : sub[n]    diag[n]  |                | right[n] |

 Operator Dx^2:   Dx^2 f[i] == minus f[i-1] + center f[i] + plus f[i+1] + coef
 off the boundary: minus = dxminus[i], center = dxcenter[i], plus = dxplus[i]

 sub[i]   = c minus
 diag[i]  = 1 + c center - Lin[i]
 sup[i]   = c plus
 right[i] = Result[i] + cc ( elem[i] center + elem[i-1] minus + elem[i+1] plus ) + (cc - c) coef
                                       
***************************************************************************************/

void FieldObj::Run1D( double c, double cc, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   ri, plus, minus, center, coef;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) );
  return;
  }

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {
    n++;
    if ( bcvar & SURF_XMIN )   istart = i; 

    (this->*nablaX)(i, i, bcvar, minus, center, plus, coef);

    ri = Result[i] + (cc - c) * coef + cc * center * elem[i];

    if ( ! (bcvar & SURF_XMIN) ) ri += cc * minus * elem[i-1];
    if ( ! (bcvar & SURF_XMAX) ) ri += cc * plus * elem[i+1]; 

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * center ;
    sup[n] = c * plus;  sub[n] = c * minus;
    right[n] = ri;
    if (++i == xsize) break;

    } // while bcvar > 0 and i < xsize
 
  if (n)  {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
  }

  if (i < xsize)
    while ( !(ptype[i] & VARY_MASK) )   {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       if ( ++i == xsize) break; 
    }

  } while (i < xsize);

}

     
/***************************************************************************************
 
                     CYLINDRICAL GEOMETRY RUN, "r" DIRECTION
                                                            
 Solving  (1 + c Dr^2 - Lin) Result_new = ( cr Dr^2 + cz Dz^2 ) This + Result
                                                          
***************************************************************************************

                     CONICAL GEOMETRY RUN, "r" DIRECTION:
                                                           1 
 Solving  (1 + c Dr^2 - Lin) Result_new = ( cr Dr^2 + cth --- Dth^2 ) This + Result
                                                          r^2
***************************************************************************************

                     POLAR GEOMETRY RUN, "r" DIRECTION:
                                                            1 
 Solving  (1 + c Dr^2 - Lin) Result_new = ( cr Dr^2 + cphi --- Dphi^2 ) This + Result
                                                           r^2
***************************************************************************************/

void FieldObj::Run2Dx( double c, double cr, double cz0, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      ix = 0, iy = 0, n = 0;
double   ri, rplus, rminus, rcenter, rcoef;
double   zplus, zminus, zcenter, zcoef, cz = cz0;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) );
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK )   {
    n++;
    if ( bcvar & SURF_XMIN )   istart = i; 
    (this->*nablaX)(i, ix, bcvar, rminus, rcenter, rplus, rcoef);     // calculating D_r^2
    (this->*nablaY)(i, iy, bcvar, zminus, zcenter, zplus, zcoef);

    if (GEOMETRY == CONICAL || GEOMETRY == POLAR ) {
      if (xcoord[ix] == 0) cz = 0 ;//cz0 * 100 / ( xgrid[ix] * xgrid[ix] );   // no nabla theta at origin
      else cz = cz0 / ( xcoord[ix] * xcoord[ix] ); 
    }

    ri = Result[i] + (cr - c) * rcoef + cz * zcoef + (cr * rcenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cr * rminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cr * rplus * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cz * zminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cz * zplus * elem[i+xsize];    

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * rcenter;
    sup[n] = c * rplus;  sub[n] = c * rminus;
    right[n] = ri; 
    ++i;
    if (++ix == xsize) break;

  } // while bcvar > 0 and i < xsize
 
  if (n)  {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
  }

  if (ix < xsize)
    while ( !(ptype[i] & VARY_MASK) )   {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       ++i;
       if ( ++ix == xsize) break; 
    }

  if (ix == xsize) {
    ix = 0;
    iy ++;
  }

} while (iy < ysize);

}

/***************************************************************************************
 
                     CYLINDRICAL GEOMETRY RUN, "z" DIRECTION
                                                            
 Solving  (1 + c Dz^2 - Lin) Result_new = ( cr Dr^2 + cz Dz^2 ) This + Result
                                                          
***************************************************************************************

                     CONICAL GEOMETRY RUN, "theta" DIRECTION:
                  1                                             1 
 Solving  (1 + c --- Dth^2 - Lin) Result_new = ( cr Dr^2 + cth --- Dth^2 ) This + Result
                 r^2                                           r^2

***************************************************************************************

                     POLAR GEOMETRY RUN, "phi" DIRECTION:
                  1                                               1 
 Solving  (1 + c --- Dphi^2 - Lin) Result_new = ( cr Dr^2 + cphi --- Dphi^2 ) This + Result
                 r^2                                             r^2
***************************************************************************************/

void FieldObj::Run2Dy( double c0, double cr, double cz0, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      ix = 0, iy = 0, n = 0, ix0=-1;
double   ri, a = 1.0, c=c0, cz=cz0;
double   rplus, rminus, rcenter, rcoef;
double   zplus, zminus, zcenter, zcoef;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) );
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {

    n++;
    if ( bcvar & SURF_YMIN )  istart = i; 

    (this->*nablaY)(i, iy, bcvar, zminus, zcenter, zplus, zcoef);
    (this->*nablaX)(i, ix, bcvar, rminus, rcenter, rplus, rcoef);     // calculating D_r^2  

    if ( (GEOMETRY == CONICAL || GEOMETRY == POLAR) && (ix0 != ix) ) {
      if (xcoord[ix] == 0) a = 0; //10 / ( xgrid[ix] * xgrid[ix] );   // no nabla theta at origin
         else a = 1 / ( xcoord[ix] * xcoord[ix] ); 
         cz = cz0 * a; c = c0 * a; ix0 = ix;
    }
   
    ri = Result[i] + cr * rcoef + (cz - c) * zcoef + (cr * rcenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) ) ri += cr * rminus * elem[i-1];
    if ( ! (bcvar & SURF_XMAX) ) ri += cr * rplus * elem[i+1];   
    if ( ! (bcvar & SURF_YMIN) ) ri += cz * zminus * elem[i-xsize];
    if ( ! (bcvar & SURF_YMAX) ) ri += cz * zplus * elem[i+xsize];   

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * zcenter;
    sup[n] = c * zplus; sub[n] = c * zminus;
    right[n] = ri; 

    i += xsize;
    if (++iy == ysize) break;

    } // while bcvar > 0 and iy < ysize
 
  if (n)
    {
      //for  (int t = 1; t <= n; t++) fprintf(stderr,"%g %g %g %g\n",sub[t],diag[t],sup[t],right[t]);
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind += xsize) Result[ind] = right[ii];
    n = 0;
    }

  if (iy < ysize)
    while ( !(ptype[i] & VARY_MASK) )  
       {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       i += xsize;
       if ( ++iy == ysize) break; 
       }

  if (iy == ysize) {
    iy = 0;
    i = ++ix;
  }

  } while (ix < xsize);

}

/***************************************************************************************

This scheme applies to ALL geometries, but does not work when both obstacles and inhomogeneous grid are present:
 
  Laplace f[i]= dplus[i] f[i+1] + dminus[i] f[i-1] - (dplus[i] + dminus[i]) f[i] 
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

 Boundary at i=0:  f[-1] = f[1] - 2 dx (a - b f[0])  where a = bc_coef + bgr b, b = bc_lin 

  dplus[0] f[1] + dminus[0] (f[1] - 2 dx (a - b f[0])) - (dplus[0] + dminus[0]) f[0] 

  (dplus[0] + dminus[0]) (f[1] - f[0]) + 2 b dx dminus[0] f[0] - 2 a dx dminus[0]  

or: 

  (dplus[0] + dminus[0]) (f[1] - f[0]) + f[0] (2 dx dxminus[0]) bc_lin - (2 dx dminus[0]) (bc_lin + bgr bc_coef)
  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

 Boundary at i=N:  f[N+1] = f[N-1] - 2 dx (a - b f[N])  where a = bc_coef + bgr b, b = bc_lin 

  dplus[N] (f[N-1] - 2 dx (a - b f[N])) + dminus[N] f[N+1] - (dplus[N] + dminus[N]) f[N]

  (dplus[N] + dminus[N]) (f[N-1] - f[N]) + 2 b dx dplus[N] f[N] - 2 a dx dplus[N]  

or: 

  (dplus[N] + dminus[N]) (f[N-1] - f[N]) + f[N] (2 dx dplus[N]) bc_lin - (2 dx dplus[N]) (bc_lin + bgr bc_coef)
  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  O L D E R    C O M M E N T S :
------------------------------------------------------------------------------------ 

 Operator Dx^2:   Dx^2 f[i] == dxminus[i] f[ix-1] + dxcenter[i] f[i] + dxplus[i] f[ix+1]
                          i == (ix,iy,iz)

 Special case of Neumann boundary ( df/dx[i] + bc_lin (f[i] - bgr) = bc_coef ):
 Assume uniform for accuracy!!

  |       |      |    f[1] - f[-1] = 2 xgrid[0] ( bc_coef + bc_lin * bgr - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 xgrid[0] ( bc_coef + bc_lin * bgr - f[0] bc_lin )
 node

  Dx^2 f[0] = ( -2 f[0] + f[1] + f[-1] ) / xgrid[0]^2
            = 2 ( f[1] - f[0] ) / xgrid[0]^2 - 2 ( bc_coef - f[0] bc_lin ) / xgrid[0]
            = 2 f[0] ( bc_lin - 1 / xgrid[0] ) / xgrid[0] +          | -> center
              2 f[1] / xgrid[0]^2 -                                  | -> right
            - 2 (bc_coef + bc_lin * bgr) / xgrid[0]                  | -> coef

=======================================================================================

"Tortuosity": D = D(r)


             ( D[i+1/2] (f[i+1] - f[i]) / xgrid[i] - D[i-1/2] (f[i] - f[i-1]) / xgrid[i-1] ) 
 Dx^2 f[i] = ------------------------------------------------------------------------------- =
                                 (xgrid[i] + xgrid[i-1])  / 2

                    D[i] + D[i-1]                               D[i+1] + D[i]
 = (f[i-1] - f[i]) -------------- dxminus[i] + (f[i+1] - f[i]) -------------- dxplus[i]
                         2                                           2


Neumann b.c.:

  Dx^2 f[0] = ( D[1/2] ( f[1] - f[0] ) - D[-1/2] ( f[0] - f[-1] ) ) / xgrid[0]^2
            = ( 2 D[0] ( f[1] - f[0] ) / xgrid[0]^2 - 2 D[-1/2] ( bc_coef + bc_lin * bgr - f[0] bc_lin ) / xgrid[0]
            = 2 f[0] ( D[-1/2] bc_lin - D[0] / xgrid[0] ) / xgrid[0] +  | -> center
              2 D[0] f[1] / xgrid[0]^2 -                                | -> plus
            - 2 D[-1/2]  (bc_coef + bc_lin * bgr) / xgrid[0]            | -> coef

  2 D[-1/2] = 3 D[0] - D[1]

...Assume D[1/2] + D[-1/2] = 2 D[0]

***************************************************************************************/

void FieldObj::nabla1Dx(long i, int ix, long bcvar,
                      double &minus, double &center, double &plus, double &coef)
{
    int bc;
    double pterm = 0;
    double Di = Diff[i];
    double Dm = 0.5 * (Di + Diff[i - (ix ? 1 : 0)] );
    double Dp = 0.5 * (Di + Diff[i + (ix + 1 < xsize ? 1 : 0)] );

    center = - (plus = Dp * dxplus[ix]) - (minus = Dm * dxminus[ix]);
    coef = 0.0;

    if ( bcvar & SURF_XMIN ) {
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)
	coef = minus;
      else 
        center = - (plus = 2 * Di * dx2inv[ix] )
                 - bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) * (coef = - 2 * Di * dxinv[ix]);

      coef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
      minus = 0;
    }
    else if ( bcvar & SURF_XMAX ) {
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0) 
	coef = plus;
      else  
        center = - (minus = 2 * Di * dx2inv[ix-1] )
                 - bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) * (coef = - 2 * Di * dxinv[ix-1]);

      coef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
      plus = 0;
    }
}

//******************* calculating D_y^2  *************************

void FieldObj::nabla1Dy(long i, int iy, long bcvar,
                      double &minus, double &center, double &plus, double &coef)
{
    int bc;
    double pterm = 0;
    double Di = Diff[i];
    double Dm = 0.5 * (Di + Diff[i - (iy > 0 ? xsize : 0)] );
    double Dp = 0.5 * (Di + Diff[i + (iy + 1 < ysize ? xsize : 0)] );

    center = - (plus = Dp * dyplus[iy]) - (minus = Dm * dyminus[iy]);
    coef = 0.0;

    if ( bcvar & SURF_YMIN ) {
      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)
	coef = minus;
      else 
        center = - (plus = 2 * Di * dy2inv[iy] )
                 - bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) * (coef = - 2 * Di * dyinv[iy]);

      coef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
      minus = 0;
    }
    else if ( bcvar & SURF_YMAX ) {
      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
	coef = plus;
      else  
        center = - (minus = 2 * Di * dy2inv[iy-1] )
                 - bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) * (coef = - 2 * Di * dyinv[iy-1]);

      coef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
      plus = 0;
    }
}

//******************* calculating D_z^2  *************************

void FieldObj::nabla1Dz(long i, int iz, long bcvar,
                      double &minus, double &center, double &plus, double &coef)
{
    int bc;
    double pterm = 0;
    double Di = Diff[i];
    double Dm = 0.5 * (Di + Diff[i - (iz > 0 ? xysize : 0)] );
    double Dp = 0.5 * (Di + Diff[i + (iz + 1 < zsize ? xysize : 0)] );

    center = - (plus = Dp * dzplus[iz]) - (minus = Dm * dzminus[iz]);
    coef = 0.0;

    if ( bcvar & SURF_ZMIN ) {
      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK ] == 0)
	coef = minus;
      else 
        center = - (plus = 2 * Di * dz2inv[iz] )
                 - bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) * (coef = -2 * Di * dzinv[iz]);

      coef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
      minus = 0;
    }
    else if ( bcvar & SURF_ZMAX ) {
      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK ] == 0) 
	coef = plus;
      else  
        center = - (minus = 2 * Di * dz2inv[iz-1] )
                 - bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) * (coef = -2 * Di * dzinv[iz-1]);

      coef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
      plus = 0;
    }
}

/***************************************************************************************
 
                          POLAR COORDINATES DIFFERENTIAL 
                          ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^                                  
 with Tortuosity:

                1  d        df
 Dr^2 f[i,j] == - --- ( D r -- ) == rminus[i] f[i-1] + rcenter[i] f[i] + rplus[i] f[i+1]
                r  dr       dr 
 where r == r[i]

  r[i+1/2] D[i+1/2] (f[i+1]-f[i]) / dr[i] - D[i-1/2] r[i-1/2] (f[i] - f[i-1]) / dr[i-1] 
= ------------------------------------------------------------------------------------- =
                           r[i]  ( dr[i] + dr[i-1] ) / 2

         2                      D[i+1/2] r[i+1/2]          D[i-1/2] r[i-1/2]         
---------------------- { f[i+1] ----------------- + f[i-1] ----------------- - f[i] (..sum..) } =
r[i] (dr[i] + dr[i-1])                dr[i]                      dr[i-1]            


 rplus[i]  = (D[i] + D[i+1]) / 2 * ( 2 / dr[i]   + 1 / r[i] ) / (dr[i] + dr[i-1] ) 
 rminus[i] = (D[i] - D[i+1]) / 2 * ( 2 / dr[i-1] - 1 / r[i] ) / (dr[i] + dr[i-1] ) 

----------------------------------

 Special case of Neumann boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform grid at boundary for accuracy!!

  |       |      |    f[1] - f[-1] = 2 xgrid[0] ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 dr ( bc_coef - f[0] bc_lin )
 node

               1                  r + dr/2                 r - dr/2         
 Dr^2 f[0] = ------ { D[1/2] f[1] -------- + D[-1/2] f[-1] -------- - f[0] ( ..sum.. ) }
             r dr                    dr                       dr            

 D[-1/2] = (3 D[0] - D[1]) / 2

   1    /               / 2 r        D[1]-D[0] \                    dr                             \
= ---- | (f[1] - f[0]) | --- D[0] + ---------   | - 2 D[-1/2] ( r - --- ) (bc_coef - bc_lin * f[0]) |
  r dr  \               \ dr             2     /                     2                             /

                   /  2          D[1]-D[0] \              2    1 
  = (f[1] - f[0]) | ----- D[0] + --------   | - D[-1/2] ( -- - - ) ( bc_coef - bc_lin * f[0] )
                   \ dr^2         2 r dr   /              dr   r

***************************************************************************************/

void FieldObj::nablaDiscR(long i, int ix, long bcvar, 
                      double &rminus, double &rcenter, double &rplus, double &rcoef)
{
  double a, b;
  int bc;
  double Di = Diff[i];
  double Dm = 0.5 * (Di + Diff[i - (ix > 0 ? 1 : 0)] );
  double Dp = 0.5 * (Di + Diff[i + (ix < xsize - 1 ? 1 : 0)] );

    rcenter = - ( (rminus = Dm * dxminus[ix]) + (rplus = Dp * dxplus[ix]) );
    rcoef = 0.0;

    if ( bcvar & SURF_XMIN )
      {
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
        rcoef = rminus * bc_coef[bc];
      else 
        if (xcoord[ix] > 0 ) { 
          a = Di * 2 * dxinv[ix];                        // a = 2 / dr
          b = a - Di / xcoord[ix];                       // b = 2/dr - 1 / r
          rplus = a * dxinv[ix];                         // rplus = 2 / dr^2
          rcenter = - rplus + b * bc_lin[bc] / (1 + elem[i] * bc_pump[bc]); // rcenter = - 2 / dr^2 + Lin (2 / dr - 1 /r)
          rcoef = - b * (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) ); // rcoef = - (2/dr - 1/r) Coef
	}
        else rcenter = - ( rplus = 4 * Di * dx2inv[ix] );
      rminus = 0;
      }
    else if ( bcvar & SURF_XMAX ) 
      {
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
         rcoef = rplus * bc_coef[bc];
      else 
        {
        a = Di * 2 * dxinv[ix-1];                       // a = 2 / dr
        b = a + Di / xcoord[ix];                        // b = 2/dr + 1 / r
        rminus = a * dxinv[ix-1];                       // rplus = 2 / dr^2
        rcenter = - rminus + b * bc_lin[bc] / (1 + elem[i] * bc_pump[bc]);             // rcenter = - 2 / dr^2 + Lin (2 / dr - 1 /r)
        rcoef = - b * (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );  // rcoef = - (2/dr + 1/r) Coef
        }
      rplus = 0;
      }
}
/***************************************************************************************

                       SPHERICAL COORDINATES DIFFERENTIAL
                       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

With tortuosity:

                 1   d          df
 Dr^2 f[i,j] == --- --- ( D r^2 -- ) == rminus[i] f[i-1] + rcenter[i] f[i] + rplus[i] f[i+1]
                r^2  dr         dr 
 where r == r[i]

       1   D[i+1/2] r[i+1/2]^2 (f[i+1]-f[i]) / dr[i] - D[i-1/2] r[i-1/2]^2 (f[i] - f[i-1]) / dr[i-1] 
 =  ------ ----------------------------------------------------------------------------------------- =
    r[i]^2                                 ( dr[i] + dr[i-1] ) / 2

           2                       r[i+1/2]^2                   r[i-1/2]^2                   
 ------------------------ { f[i+1] ---------- D[i+1/2] + f[i-1] ---------- D[i-1/2] - f[i] ( ..sum.. ) }
 r[i]^2 (dr[i] + dr[i-1])             dr[i]                       dr[i-1]                    


 rplus[i]  = (D[i] + D[i+1]) * (r[i+1/2]/r[i])^2 / dr[i] / (dr[i] + dr[i-1]) 
 rminus[i] = (D[i] + D[i-1]) * (r[i-1/2]/r[i])^2 / dr[i-1] / (dr[i] + dr[i-1]) 

----------------------------------

 Special case of Neumann boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform grid at boundary for accuracy!!

  |       |      |    f[1] - f[-1] = 2 dr ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 dr ( bc_coef - f[0] bc_lin )
 node

              1               
Dr^2 f[0] = ------ { f[1] D[1/2] rplus[0] + f[-1] D[-1/2] rminus[0] - f[0] ( ..sum.. ) }
            r^2 dr            

      1                           4 r^2               2 r
  = ----- { ( f[1] - f[0] ) ( 1 + ------ ) - dr ( 1 - ---- )^2 ( bc_coef - bc_lin * f[0] ) }
    2 r^2                          dr^2                dr

-----------------------------------

  |       |      |      f[-1] - f[1] = 2 dr ( bc_coef - f[0] bc_lin )
 f[-1]  f[0]  virtual   Eliminating f[1]: f[1] = f[-1] - 2 dr ( bc_coef - f[0] bc_lin )
               node

               1           (r + dr/2)^2         (r - dr/2)^2          (r + dr/2)^2   (r - dr/2)^2
 Dr^2 f[0] = ------ { f[1] ------------ + f[-1] ------------ - f[0] ( ------------ + ------------ ) }
             r^2 dr             dr                  dr                     dr             dr

      1                            4 r^2               2 r
  = ----- { ( f[-1] - f[0] ) ( 1 + ------ ) - dr ( 1 + ---- )^2 ( bc_coef - bc_lin * f[0] ) }
    2 r^2                           dr^2                dr



====================================
                                
Saturating pump: df/dt = V f / (f + K) --> df/dr (f + K) - V f / Ca.D = 0
      1
  d ( - f*f + Kf ) / dr - v f = 0 , where v == V / Ca.D
      2

 Linearization: f[n+1] f[n+1] = f[n] f[n] + 2 f[n] (f[n+1] - f[n]) = 2 f[n] f[n+1] - f[n] f[n]
               1
  d ( fOld f - - fOld^2 + K f) / dr - v f = 0 -- boundary condition for f[n+1]
               2
      1 
  d ( - fOld^2 + K fOld ) / dr - v fOld = 0  -- boundary condition for f[n]
      2

  fOld[1] f[1] - fOld[-1] f[-1] - .5 fOld[1]^2 + .5 fOld[-1]^2 + K f[1] - K f[-1] = dr v f[0] 
                                
  

***************************************************************************************/

// rplus[i]  = (D[i] + D[i+1]) * (r[i+1/2]/r[i])^2 / dr[i] / (dr[i] + dr[i-1]) 
// rminus[i] = (D[i] + D[i-1]) * (r[i-1/2]/r[i])^2 / dr[i-1] / (dr[i] + dr[i-1]) 

/***************************************************************************************/

void FieldObj::nablaSphereR(long i, int ix, long bcvar, 
                      double &rminus, double &rcenter, double &rplus, double &rcoef)
{
  double a, b;
  int bc;
  double Di = Diff[i];
  double Dm = 0.5 * (Di + Diff[i - (ix > 0 ? 1 : 0)] );
  double Dp = 0.5 * (Di + Diff[i + (ix + 1 < xsize ? 1 : 0)] );

    rcenter = - (rminus = Dm * dxminus[ix]) - (rplus = Dp * dxplus[ix]);
    rcoef = 0;

    if ( bcvar & SURF_XMIN )
      {
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
        rcoef = rminus * bc_coef[bc];
      else 
        if (xcoord[ix] > 0 ) { 
          a = 2 * xcoord[ix] * dxinv[ix];
          b = Di * 0.5 / ( xcoord[ix] * xcoord[ix] );
          rplus = b * (1 + a * a );
          rcoef = - b * (1 - a) * (1 - a) * xgrid[ix]; 
          rcenter = - (rplus + rcoef * bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) );
          rcoef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
 	  }
        else rcenter = - ( rplus = 6 * Di * dx2inv[ix] );
      rminus = 0;
      }
    else if ( bcvar & SURF_XMAX ) 
      {
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
         rcoef = rplus * bc_coef[bc];
      else 
        {
        a = 2 * xcoord[ix] * dxinv[ix-1];
        b = 0.5 * Di / ( xcoord[ix] * xcoord[ix] );
        rminus = b * ( 1 + a * a );
        rcoef = - b * (1 + a) * (1 + a) * xgrid[ix-1]; 
        rcenter = - (rminus + rcoef * bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) );
        rcoef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
        }
      rplus = 0;
      }
}


/***************************************************************************************

                        SPHERICAL COORDINATES THETA DIFFERENTIAL
                        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

                         1       d              df
 Dth^2 f[j,i] == -------------- ---- ( sin(th) ---- ) == dyminus[i] f[i-1] + dycenter[i] f[i] + dyplus[i] f[i+1]
                 r[j]^2 sin(th) d th           d th

 where th == theta[i]
 
        1      sin(th[i+1/2]) (f[i+1] - f[i]) / dth[i] - sin(th[i-1/2]) (f[i] - f[i-1]) / dth[i-1]
 = ----------- ---------------------------------------------------------------------------------
   r^2 sin(th)                               (dth[i] + dth[i-1]) / 2



 At Neumann boundary:  f[j-1] = f[j+1] - 2 dygrid[j] (bc_coef - bc_lin f[j]),

                       dth[i] = dth[i-1] == dth

                 2          / (f[j+1] - f[j]) sin(theta) cos(dth/2)                                        \
 Dth^2 f = --------------- |  ------------------------------------- - sin(th-dth/2) (bc_coef - bc_lin f[j]) |
           r^2 sin(th) dth  \                 dth                                                          /

  Note that the "sin(theta)" cancels in the first term


 ***************************************************************************************/



void FieldObj::nablaTheta(long i, int iy, long bcvar, 
                      double &thminus, double &thcenter, double &thplus, double &thcoef)
{
  int bc;
  double Di = Diff[i];
  double Dm = 0.5 * (Di + Diff[i - (iy > 0 ? xsize : 0)] );
  double Dp = 0.5 * (Di + Diff[i + (iy < ysize - 1 ? xsize : 0)] );


    thcenter = - (thminus = Dm * dyminus[iy]) - (thplus = Dp * dyplus[iy]);
    thcoef = 0.0;

    if (bcvar & SURF_YMIN) {
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)  
          thcoef = thminus * bc_coef[bc];
        else 
          if (ycoord[iy] > 0) { 
            thplus = 2 * Di * cos( 0.5 * ygrid[iy] ) * dy2inv[iy];
            thcoef = - 2 * Di * dyinv[iy] * sin( ycoord[iy] - 0.5 * ygrid[iy]) / sin( ycoord[iy] ) ;
            thcenter = - (thplus + thcoef * bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) );
            thcoef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]));
 	    }
	  else    // theta = 0 
            thcenter = - (thplus = 4 * Di * dy2inv[iy]);
	thminus = 0;
    }
    else  if (bcvar & SURF_YMAX) {
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)  
          thcoef = thplus * bc_coef[bc];
        else 
          if ( fabs( ycoord[iy] - pi) > 1.0e-12 ) { 
            thminus = 2 * Di * cos( 0.5 * ygrid[iy-1] ) * dy2inv[iy-1];
            thcoef = - 2 * Di * dyinv[iy-1] * sin( ycoord[iy] + 0.5 * ygrid[iy-1]) / sin( ycoord[iy] ) ;
            thcenter = - (thminus + thcoef * bc_lin[bc] / (1 + elem[i] * bc_pump[bc]) );
            thcoef *= (bc_coef[bc] + bgr * bc_lin[bc] / (1 + bgr * bc_pump[bc]) );
 	    }
	  else   // theta =  pi 
            thcenter = - (thminus = 4 * Di * dy2inv[iy-1]);
        thplus = 0.0;
    }

}

//**********************************************************************************************
/*

 void FieldObj::cleanBoundaries()
    { 
    for (long i = 0; i < size; i++)  
      if ( !(ptype[i] & VARY_MASK) )
        elem[i] = (ptype[i] & DIRICHLET) ? bc_coef[ptype[i] >> XSHIFT] : 0;
    }
*/
//***************************************************************

void FieldObj::print_ptype(long i)
{
long x = ptype[i];
if (x & _OUTSIDE_) fprintf(stdlog,"[OUT] ");
else
  {
  if (x & VARY_MASK) fprintf(stdlog,"[VARY");
  else if (x & DIRICHLET) fprintf(stdlog,"[f=%g", bc_coef[x >> XSHIFT]); 
  if (x & SURF_MASK)
    {
    if (x & SURF_XMIN) fprintf(stdlog," %s XMIN", bc_id[(x >> XSHIFT) & BC_ID_MASK]);
    if (x & SURF_XMAX) fprintf(stdlog," %s XMAX", bc_id[(x >> XSHIFT) & BC_ID_MASK]);
    if (x & SURF_YMIN) fprintf(stdlog," %s YMIN", bc_id[(x >> YSHIFT) & BC_ID_MASK]);
    if (x & SURF_YMAX) fprintf(stdlog," %s YMAX", bc_id[(x >> YSHIFT) & BC_ID_MASK]);
    if (x & SURF_ZMIN) fprintf(stdlog," %s ZMIN", bc_id[(x >> ZSHIFT) & BC_ID_MASK]);
    if (x & SURF_ZMAX) fprintf(stdlog," %s ZMAX", bc_id[(x >> ZSHIFT) & BC_ID_MASK]);
    }
   fprintf(stdlog,"] ");
  }
}


//****************************************************************************
//*                   C L A S S   B U F F E R   A R R A Y
//****************************************************************************

  BufferObj :: BufferObj(TokenString &params, const char *id) : 
  FieldObj(params, id)
    {
    KD = kplus = kminus = -1.0;

    params.get2_param(id, "koff",  &kminus);
    params.get2_param(id, "kminus",&kminus);
    params.get2_param(id, "kon",   &kplus); 
    params.get2_param(id, "kplus", &kplus); 
    params.get2_param(id, "KD",    &KD);  

    if (kplus > 0 && kminus >= 0)   KD = kminus / kplus;
    else if (kplus >= 0 && KD >= 0) kminus = kplus * KD;
    else if (kminus >= 0 && KD > 0) kplus  = kminus / KD;

    if ( fabs(KD * kplus - kminus) > 1e-9 || KD * kplus * kminus == 0.0 ) 
         throw makeMessage("Incompatible properties for buffer \"%s\": on rate = %g, off rate = %g, KD = %g",
                            ID, kplus, kminus, KD);
    if (VERBOSE) 
      fprintf(stdlog,"    Kinetics for buffer %s: KD=%g uM, k-=%g /ms, k+=%g /(ms uM)\n",
	             ID, KD, kminus, kplus);

   
    total = new VectorObj(Size);
    double Total = 0;
    long pos, ii, pMembrane = 0;

    if (VERBOSE) fprintf(stdlog, "    Total [%s] = ", ID);

    if ( !params.token_count(id, "total", &pos) ) 
      throw makeMessage("Total concentration not defined for buffer \"%s\"", id);

    params.token3_count("buffer", ID, "membrane", &pMembrane);

    if ( D != 0.0 || pMembrane ) 
      Total = ExpressionObj(params, pos + 1, makeMessage("Bad total concentration definition for buffer \"%s\"", id) ).Evaluate();
    if ( D != 0.0 ) {
      if (VERBOSE) fprintf(stdlog,"%g uM", Total);
      *total = Total;
    }

    if ( pMembrane ) {
        if (D > 0.0) 
           params.errorMessage(params.token_index(ID, "D") + 1, 0, "Diffusion coefficient must be zero for membrane-bound buffer");
        double  Depth = 0;
        params.trail_pars( pMembrane, 'd', &Depth);
        if (Depth == 0.0)  setMembraneBound(Total);
                    else   setMembraneLayer(Total, Depth);
      }
    else if (D == 0.0) 
            setFieldByFunction(params, pos + 1, total->elem, 
                               makeMessage("Bad expression for total concentration of %s",ID) );

   if (!params.token_count(id, "import")) {

      double calcium = 0;
      calcium = params.get_param("Ca.bgr", 0, "Undefined background [Ca]");
      for (ii=0 ; ii < Size; ii++) {
         (*this)[ii] = (*total)[ii]  / (1.0 + kplus / kminus * calcium);
         if ((*total)[ii] < 0 || (*this)[ii] < 0 )
             params.errorMessage( pos+1, makeMessage("Negative concentration for buffer %s at node %ld", ID, ii) );
      }
    }

    bgr = average();
    if ( VERBOSE ) fprintf(stdlog,"\n    Free background [%s] = %g uM\n", ID, bgr);


    //printf("Initialize total in %s\n",ID);
    for (long ind = 0; ind < Size; ind++) if ( !(ptype[ind] & VARY_MASK) ) (*total)[ind] = (*this)[ind] = 0.0; 
   // initialAve = bgr;
    //printf("Done\n");
    }

//****************************************************************************

  void  BufferObj :: setMembraneLayer(double Total, double depth) 
    {
    long    ii, jj, Incr;
    int     ix, iy, iz, j, incr;
    double  *coord, dx;

    *total = 0.0;

    if (VERBOSE) fprintf(stdlog,"%g uM\n    Buffer %s is membrane-localized with depth %g um", Total, ID, depth); 
    
    for (ii=0 ; ii < Size; ii++) 
     if ( ptype[ii] & SURF_MASK ) { 

       Grid->split(ii, ix, iy, iz);
       unsigned long point = ptype[ii];

       if ( point & (SURF_XMAX | SURF_XMIN) ) 
          { Incr = 1;      dx = xgrid[j = ix]; coord = xcoord; incr = (point & SURF_XMIN) ? 1 : -1; }
       if ( point & (SURF_YMAX | SURF_YMIN) ) 
          { Incr = xsize;  dx = ygrid[j = iy]; coord = ycoord; incr = (point & SURF_YMIN) ? 1 : -1; }
       if ( point & (SURF_ZMAX | SURF_ZMIN) ) 
          { Incr = xysize; dx = zgrid[j = iz]; coord = zcoord; incr = (point & SURF_ZMIN) ? 1 : -1; }

       Incr *= incr;

       jj = ii;

       //fprintf(stderr,"ii=%ld (%d,%d,%d) point=%ld Incr=%ld incr=%d dx=%g\n",ii,ix,iy,iz,point,Incr,incr,dx); 

       double coordinate = coord[j];
       do {
         if (ptype[jj] & VARY_MASK )  
         total->elem[jj] = Total; 
         jj += Incr; j += incr; 
       } while ( fabs(coord[j] - coordinate) <= depth && jj < Size );
   }    
 }

//****************************************************************************

  void  BufferObj :: setMembraneBound(double Total) 
    {
    long    ii;
    int     ix, iy, iz;
    double  Avogadro = 602.2;  // Units of concentration per area are:
                               // uM * um = 10^(-13) mol /cm^2 = 10^(-21) mol / um^2 = 602.2 1 / um^2

    *total = 0.0;

    if (VERBOSE) fprintf(stdlog,"%g / um^2\n    Buffer %s is membrane-bound", Total, ID);
    Total /= Avogadro; 

    for (ii=0 ; ii < Size; ii++) 
     if ( ptype[ii] & SURF_MASK ) { 

       Grid->split(ii, ix, iy, iz);
       unsigned long point = ptype[ii];

       if ( point & (SURF_XMAX | SURF_XMIN) ) total->elem[ii] += Total / xgrid[ix];
       if ( point & (SURF_YMAX | SURF_YMIN) ) total->elem[ii] += Total / ygrid[iy];
       if ( point & (SURF_ZMAX | SURF_ZMIN) ) total->elem[ii] += Total / zgrid[iz];
   }    
 }

//**************************************************************************



BufferArray::BufferArray(TokenString &params, const char *group) {
    
    ID = StrCpy( group );
    buf_num = params.token_count(group);
    array = new BufferObj *[buf_num];

    if ( !buf_num )  return;

    for (int i=0; i<buf_num; i++) {
      long p = params.token_index(ID, i+1) + 1;
      char *id = params.getVarName( p, group );
      for (int j = i-1; j>0; j--) 
        if ( equal( id, array[j]->ID) ) 
           params.errorMessage(p+1, makeMessage("Buffer \"%s\" defined more than once",ID) );
      array[i] = new BufferObj(params, id);
      delete [] id;
    }
}

//****************************************************************************

void BufferArray::allocateCopy(const BufferArray &bufs)
    {
    if (bufs.ID) ID = StrCpy( bufs.ID );
    if ( (buf_num = bufs.buf_num) > 0 )
      {
      array = new BufferObj *[buf_num];
      for (int i = 0; i < buf_num; i++) array[i] = new BufferObj(*bufs.array[i]);
      }
    }

//****************************************************************************

double calcium_gain( FieldObj *Ca, BufferArray *Bufs) {
  double x = 0.0;
  if (Ca) x = Ca->gain();
  if (Bufs) x -= Bufs->gain();
  return x;
}

//****************************************************************************
